@@ -4,6 +4,7 @@
 # Borrows heavily from the Python bindings for the Stripe API
 
 # Imports
+from __future__ import print_function
 import logging
 import os
 import platform
@@ -15,12 +16,12 @@ import datetime
 import types
 import base64
 import pkg_resources
-
+import six
 # Use cStringIO if ita's available.  Otherwise, StringIO is fine.
 try:
   import cStringIO as StringIO
 except ImportError:
-  import StringIO
+  from six import StringIO
 
 # - Requests is the preferred HTTP library
 # - Google App Engine has urlfetch
@@ -50,8 +51,8 @@ if not _httplib:
     pass
   else:
     if major == 0 and (minor < 8 or (minor == 8 and patch < 8)):
-      print >>sys.stderr, 'Warning: the Clever library requires that your Python "requests" library has a version no older than 0.8.8, but your "requests" library has version %s. Clever will fall back to an alternate HTTP library, so everything should work, though we recommend upgrading your "requests" library. If you have any questions, please contact support@clever.com. (HINT: running "pip install -U requests" should upgrade your requests library to the latest version.)' % (
-          version, )
+      print('Warning: the Clever library requires that your Python "requests" library has a version no older than 0.8.8, but your "requests" library has version %s. Clever will fall back to an alternate HTTP library, so everything should work, though we recommend upgrading your "requests" library. (HINT: running "pip install -U requests" should upgrade your requests library to the latest version.)' % (
+          version, ), file=sys.stderr)
       _httplib = None
 
 if not _httplib:
@@ -65,22 +66,21 @@ if not _httplib:
   try:
     import urllib2
     _httplib = 'urllib2'
-    print >>sys.stderr, "Warning: the Clever library is falling back to urllib2 because pycurl isn't installed. urllib2's SSL implementation doesn't verify server certificates. For improved security, we suggest installing pycurl."
+    print("Warning: the Clever library is falling back to urllib2 because pycurl isn't installed. urllib2's SSL implementation doesn't verify server certificates. For improved security, we suggest installing pycurl.", file=sys.stderr)
   except ImportError:
     pass
 
 if not _httplib:
   raise ImportError(
-      "Clever requires one of pycurl, Google App Engine's urlfetch, or urllib2.  If you are on a platform where none of these libraries are available, please let us know at support@clever.com.")
+      "Clever requires one of pycurl, Google App Engine's urlfetch, or urllib2.")
 
-from version import VERSION
-import importer
+from .version import VERSION
+from . import importer
 json = importer.import_json()
 
 logger = logging.getLogger('clever')
 
-# Use certs chain bundle including in the package for SSL verification
-CLEVER_CERTS = pkg_resources.resource_filename(__name__, 'data/clever.com_ca_bundle.crt')
+API_VERSION = "v1.2"
 
 # Configuration variables
 
@@ -98,16 +98,6 @@ def get_token():
   global global_auth
   return global_auth.get('token', None)
 
-
-def set_api_key(api_key):
-  global global_auth
-  global_auth = {'api_key': api_key}
-
-
-def get_api_key():
-  global global_auth
-  return global_auth.get('api_key', None)
-
 # Exceptions
 
 
@@ -150,7 +140,7 @@ def convert_to_clever_object(klass, resp, auth):
       return [convert_to_clever_object(klass, i, auth) for i in resp['data']]
     elif isinstance(resp['data'], dict):
       return klass.construct_from(resp['data'].copy(), auth)
-  elif isinstance(resp, basestring) or isinstance(resp, list) or isinstance(resp, dict) or isinstance(resp, bool):
+  elif isinstance(resp, six.string_types) or isinstance(resp, list) or isinstance(resp, dict) or isinstance(resp, bool):
     return resp
   else:
     raise Exception('DONT KNOW WHAT TO DO WITH {0}'.format(resp))
@@ -192,7 +182,7 @@ class APIRequestor(object):
       return d.id
     elif isinstance(d, dict):
       res = {}
-      for k, v in d.iteritems():
+      for k, v in six.iteritems(d):
         res[k] = cls._objects_to_ids(v)
       return res
     else:
@@ -204,7 +194,7 @@ class APIRequestor(object):
     Internal: encode a dict for url representation
     If we ever need fancy encoding of embedded objects do it here
     """
-    return urllib.urlencode(d)
+    return six.moves.urllib.parse.urlencode(d)
 
   @classmethod
   def jsonencode(cls, d):
@@ -242,9 +232,9 @@ class APIRequestor(object):
     my_auth = self._auth or global_auth
     if my_auth is None:
       raise AuthenticationError(
-          'No authentication method provided. (HINT: "clever.api_key = <API-KEY>" is deprecated. Set your API key using "clever.set_api_key(<API-KEY>)" or "clever.set_token(<OAUTH-TOKEN>)". You can generate API keys from the Clever web interface.  See https://clever.com/developers/docs for details, or email support@clever.com if you have any questions.')
-    if my_auth.get('token') is None and my_auth.get('api_key') is None:
-      raise AuthenticationError('Must provide either api_key or token auth. {}'.format(my_auth))
+          'No authentication method provided. (HINT: "clever.api_key = <API-KEY>" is deprecated. Set your API token using "clever.set_token(<OAUTH-TOKEN>)". You can generate API tokens from the Clever web interface.  See https://clever.com/developers/docs for details.')
+    if my_auth.get('token') is None:
+      raise AuthenticationError('Must provide api token auth. {}'.format(my_auth))
 
     abs_url = self.api_url(url)
     params = params.copy()
@@ -260,13 +250,13 @@ class APIRequestor(object):
                        ['uname', lambda: ' '.join(platform.uname())]]:
       try:
         val = func()
-      except Exception, e:
+      except Exception as e:
         val = "!! %s" % e
       ua[attr] = val
 
     headers = {
         'X-Clever-Client-User-Agent': json.dumps(ua),
-        'User-Agent': 'Clever/v1.1 PythonBindings/%s' % (VERSION, )
+        'User-Agent': 'Clever/%s PythonBindings/%s' % (API_VERSION, VERSION)
     }
     if my_auth.get('api_key', None) != None:
       headers['Authorization'] = 'Basic {}'.format(base64.b64encode(my_auth['api_key'] + ':'))
@@ -282,7 +272,7 @@ class APIRequestor(object):
       res = make_request[_httplib](meth, abs_url, headers, params)
     else:
       raise CleverError(
-          "Clever Python library bug discovered: invalid httplib %s.  Please report to support@clever.com" % (_httplib, ))
+          "Clever Python library bug discovered: invalid httplib %s.  Please report to tech-support@clever.com" % (_httplib, ))
     logger.debug('API request to %s returned (response code, response body) of (%d, %r)' %
                  (abs_url, res['code'], res['body']))
     return res, my_auth
@@ -290,7 +280,7 @@ class APIRequestor(object):
   def interpret_response(self, http_res):
     rbody, rcode= http_res['body'], http_res['code']
     try:
-      resp = json.loads(rbody) if rcode != 429 else {'error': 'Too Many Requests'}
+      resp = json.loads(rbody.decode('utf-8')) if rcode != 429 else {'error': 'Too Many Requests'}
     except Exception:
       raise APIError("Invalid response body from API: %s (HTTP response code was %d)" %
                      (rbody, rcode), rbody, rcode)
@@ -309,22 +299,13 @@ class APIRequestor(object):
       headers['Content-Type'] = 'application/json'
     else:
       raise APIConnectionError(
-          'Unrecognized HTTP method %r.  This may indicate a bug in the Clever bindings.  Please contact support@clever.com for assistance.' % (meth, ))
+          'Unrecognized HTTP method %r.  This may indicate a bug in the Clever bindings.  Please contact tech-support@clever.com for assistance.' % (meth, ))
 
     try:
       try:
-        # Use a CA_BUNDLE containing the following chain:
-        # - TrustedRoot
-        # - DigiCert High Assurance EV - 1
-        #
-        # This ensures that only this certificate chain is used to verify SSL certs.
-        # Certs dervived from other ca certs will be treated as invalid.
-        # eg. https://api.twitter.com and https://api.stripe.com FAIL
-        #     https://api.clever.com and https://api.github.com PASS
         result = requests.request(meth, abs_url,
-                                  headers=headers, data=data, timeout=80,
-                                  verify=CLEVER_CERTS)
-      except TypeError, e:
+                                  headers=headers, data=data, timeout=80)
+      except TypeError as e:
         raise TypeError(
             'Warning: It looks like your installed version of the "requests" library is not compatible with Clever\'s usage thereof. (HINT: The most likely cause is that your "requests" library is out of date. You can fix that by running "pip install -U requests".) The underlying error was: %s' % (e, ))
 
@@ -334,7 +315,7 @@ class APIRequestor(object):
       content = result.content
       status_code = result.status_code
       headers = result.headers
-    except Exception, e:
+    except Exception as e:
       # Would catch just requests.exceptions.RequestException, but can
       # also raise ValueError, RuntimeError, etc.
       self.handle_requests_error(e)
@@ -342,10 +323,10 @@ class APIRequestor(object):
 
   def handle_requests_error(self, e):
     if isinstance(e, requests.exceptions.RequestException):
-      msg = "Unexpected error communicating with Clever.  If this problem persists, let us know at support@clever.com."
+      msg = "Unexpected error communicating with Clever.  If this problem persists, let us know at tech-support@clever.com."
       err = "%s: %s" % (type(e).__name__, e.message)
     else:
-      msg = "Unexpected error communicating with Clever.  It looks like there's probably a configuration issue locally.  If this problem persists, let us know at support@clever.com."
+      msg = "Unexpected error communicating with Clever.  It looks like there's probably a configuration issue locally.  If this problem persists, let us know at tech-support@clever.com."
       err = "A %s was raised" % (type(e).__name__, )
       if e.message:
         err += " with error message %s" % (e.message, )
@@ -355,8 +336,8 @@ class APIRequestor(object):
     raise APIConnectionError(msg)
 
   def pycurl_request(self, meth, abs_url, headers, params):
-    s = StringIO.StringIO()
-    rheader = StringIO.StringIO()
+    s = StringIO()
+    rheader = StringIO()
     curl = pycurl.Curl()
 
     meth = meth.lower()
@@ -375,7 +356,7 @@ class APIRequestor(object):
         raise APIConnectionError("Did not expect params in DELETE request")
     else:
       raise APIConnectionError(
-          'Unrecognized HTTP method %r.  This may indicate a bug in the Clever bindings.  Please contact support@clever.com for assistance.' % (meth, ))
+          'Unrecognized HTTP method %r.  This may indicate a bug in the Clever bindings.  Please contact tech-support@clever.com for assistance.' % (meth, ))
 
     # pycurl doesn't like unicode URLs
     abs_url = self._utf8(abs_url)
@@ -384,16 +365,12 @@ class APIRequestor(object):
     curl.setopt(pycurl.NOSIGNAL, 1)
     curl.setopt(pycurl.CONNECTTIMEOUT, 30)
     curl.setopt(pycurl.TIMEOUT, 80)
-    curl.setopt(pycurl.HTTPHEADER, ['%s: %s' % (k, v) for k, v in headers.iteritems()])
+    curl.setopt(pycurl.HTTPHEADER, ['%s: %s' % (k, v) for k, v in six.iteritems(headers)])
     curl.setopt(pycurl.HEADERFUNCTION, rheader.write)
-    if verify_ssl_certs:
-      curl.setopt(pycurl.CAINFO, CLEVER_CERTS)
-    else:
-      curl.setopt(pycurl.SSL_VERIFYHOST, False)
 
     try:
       curl.perform()
-    except pycurl.error, e:
+    except pycurl.error as e:
       self.handle_pycurl_error(e)
     return {'body': s.getvalue(), 'headers': rheader.getvalue(), 'code': curl.getinfo(pycurl.RESPONSE_CODE)}
 
@@ -401,13 +378,13 @@ class APIRequestor(object):
     if e[0] in [pycurl.E_COULDNT_CONNECT,
                 pycurl.E_COULDNT_RESOLVE_HOST,
                 pycurl.E_OPERATION_TIMEOUTED]:
-      msg = "Could not connect to Clever (%s).  Please check your internet connection and try again.  If this problem persists, you should check Clever's service status at http://status.clever.com, or let us know at support@clever.com." % (
+      msg = "Could not connect to Clever (%s).  Please check your internet connection and try again.  If this problem persists, you should check Clever's service status at http://status.clever.com." % (
           api_base, )
     elif e[0] == pycurl.E_SSL_CACERT or e[0] == pycurl.E_SSL_PEER_CERTIFICATE:
-      msg = "Could not verify Clever's SSL certificate.  Please make sure that your network is not intercepting certificates.  (Try going to %s in your browser.)  If this problem persists, let us know at support@clever.com." % (
+      msg = "Could not verify Clever's SSL certificate.  Please make sure that your network is not intercepting certificates.  (Try going to %s in your browser)." % (
           api_base, )
     else:
-      msg = "Unexpected error communicating with Clever.  If this problem persists, let us know at support@clever.com."
+      msg = "Unexpected error communicating with Clever.  If this problem persists, let us know at tech-support@clever.com."
     msg = textwrap.fill(msg) + "\n\n(Network error: " + e[1] + ")"
     raise APIConnectionError(msg)
 
@@ -423,13 +400,10 @@ class APIRequestor(object):
         raise APIConnectionError("Did not expect params in DELETE request")
     else:
       raise APIConnectionError(
-          'Unrecognized HTTP method %r.  This may indicate a bug in the Clever bindings.  Please contact support@clever.com for assistance.' % (meth, ))
+          'Unrecognized HTTP method %r.  This may indicate a bug in the Clever bindings.  Please contact tech-support@clever.com for assistance.' % (meth, ))
     args['url'] = abs_url
     args['method'] = meth
     args['headers'] = headers
-    # Google App Engine doesn't let us specify our own cert bundle.
-    # However, that's ok because the CA bundle they use recognizes
-    # api.clever.com.
     args['validate_certificate'] = verify_ssl_certs
     # GAE requests time out after 60 seconds, so make sure we leave
     # some time for the application to handle a slow Clever
@@ -437,20 +411,20 @@ class APIRequestor(object):
 
     try:
       result = urlfetch.fetch(**args)
-    except urlfetch.Error, e:
+    except urlfetch.Error as e:
       self.handle_urlfetch_error(e, abs_url)
     return {'body': result.content, 'headers': result.headers, 'code': result.status_code}
 
   def handle_urlfetch_error(self, e, abs_url):
     if isinstance(e, urlfetch.InvalidURLError):
-      msg = "The Clever library attempted to fetch an invalid URL (%r).  This is likely due to a bug in the Clever Python bindings.  Please let us know at support@clever.com." % (
+      msg = "The Clever library attempted to fetch an invalid URL (%r).  This is likely due to a bug in the Clever Python bindings.  Please let us know at tech-support@clever.com." % (
           abs_url, )
     elif isinstance(e, urlfetch.DownloadError):
       msg = "There were a problem retrieving data from Clever."
     elif isinstance(e, urlfetch.ResponseTooLargeError):
-      msg = "There was a problem receiving all of your data from Clever.  This is likely due to a bug in Clever.  Please let us know at support@clever.com."
+      msg = "There was a problem receiving all of your data from Clever.  This is likely due to a bug in Clever.  Please let us know at tech-support@clever.com."
     else:
-      msg = "Unexpected error communicating with Clever.  If this problem persists, let us know at support@clever.com."
+      msg = "Unexpected error communicating with Clever.  If this problem persists, let us know at tech-support@clever.com."
     msg = textwrap.fill(msg) + "\n\n(Network error: " + str(e) + ")"
     raise APIConnectionError(msg)
 
@@ -472,23 +446,23 @@ class APIRequestor(object):
         raise APIConnectionError("Did not expect params in DELETE request")
     else:
       raise APIConnectionError(
-          'Unrecognized HTTP method %r.  This may indicate a bug in the Clever bindings.  Please contact support@clever.com for assistance.' % (meth, ))
+          'Unrecognized HTTP method %r.  This may indicate a bug in the Clever bindings.  Please contact tech-support@clever.com for assistance.' % (meth, ))
 
     try:
       response = urllib2.urlopen(req)
       rbody = response.read()
       rheader = response.info()
       rcode = response.code
-    except urllib2.HTTPError, e:
+    except urllib2.HTTPError as e:
       rcode = e.code
       rheader = None
       rbody = e.read()
-    except (urllib2.URLError, ValueError), e:
+    except (urllib2.URLError, ValueError) as e:
       self.handle_urllib2_error(e, abs_url)
     return {'body': rbody, 'headers': rheader, 'code': rcode}
 
   def handle_urllib2_error(self, e, abs_url):
-    msg = "Unexpected error communicating with Clever.  If this problem persists, let us know at support@clever.com."
+    msg = "Unexpected error communicating with Clever.  If this problem persists, let us know at tech-support@clever.com."
     msg = textwrap.fill(msg) + "\n\n(Network error: " + str(e) + ")"
     raise APIConnectionError(msg)
 
@@ -606,7 +580,7 @@ class CleverObject(object):
       self._transient_values.add(k)
       self._unsaved_values.discard(k)
 
-    for k, v in values.iteritems():
+    for k, v in six.iteritems(values):
       if k in self._permanent_attributes:
         continue
       self.__dict__[k] = convert_to_clever_object(self, v, auth)
@@ -616,7 +590,7 @@ class CleverObject(object):
 
   def __repr__(self):
     id_string = ''
-    if isinstance(self.get('id'), basestring):
+    if isinstance(self.get('id'), six.string_types):
       id_string = ' id=%s' % self.get('id').encode('utf8')
 
     return '<%s%s at %s> JSON: %s' % (type(self).__name__, id_string, hex(id(self)), json.dumps(self.to_dict(), sort_keys=True, indent=2, cls=CleverObjectEncoder))
@@ -674,12 +648,12 @@ class APIResource(CleverObject):
     if cls == APIResource:
       raise NotImplementedError(
           'APIResource is an abstract class.  You should perform actions on its subclasses (Charge, Customer, etc.)')
-    return "%s" % urllib.quote_plus(cls.__name__.lower())
+    return "%s" % six.moves.urllib.parse.quote_plus(cls.__name__.lower())
 
   @classmethod
   def class_url(cls):
     cls_name = cls.class_name()
-    return "/v1.1/%ss" % cls_name
+    return "/%s/%ss" % (API_VERSION, cls_name)
 
   def instance_url(self):
     id = self.get('id')
@@ -688,13 +662,21 @@ class APIResource(CleverObject):
           'Could not determine which URL to request: %s instance has invalid ID: %r' % (type(self).__name__, id), 'id')
     id = APIRequestor._utf8(id)
     base = self.class_url()
-    extn = urllib.quote_plus(id)
+    extn = six.moves.urllib.parse.quote_plus(id)
     return "%s/%s" % (base, extn)
 
 # Classes of API operations
 
 
+def get_link(response, rel):
+  links = response.get('links', None)
+  for link in links:
+    if link.get('rel', None) == rel:
+      return link.get('uri')
+  return None
+
 class ListableAPIResource(APIResource):
+  ITER_LIMIT = 1000
 
   @classmethod
   def all(cls, auth=None, **params):
@@ -709,15 +691,19 @@ class ListableAPIResource(APIResource):
 
     requestor = APIRequestor(auth)
     url = cls.class_url()
-    params['page'] = 1
-    params['limit'] = 1000
-    done_paginating = False
-    while not done_paginating:
+    params['limit'] = cls.ITER_LIMIT
+
+    while url:
       response, auth = requestor.request('get', url, params)
+      # Generate nothing when there is no data.
+      if len(response['data']) == 0:
+        break
       for datum in convert_to_clever_object(cls, response, auth):
         yield datum
-      params['page'] += 1
-      done_paginating = len(response['data']) < params['limit']
+
+      url = get_link(response, 'prev' if 'ending_before' in params else 'next')
+      # params already included in url from get_link
+      params = {}
 
 
 class CreatableAPIResource(APIResource):
@@ -758,14 +744,30 @@ class DeletableAPIResource(APIResource):
 # API objects
 
 
+class Contact(ListableAPIResource):
+  pass
+
+
 class District(ListableAPIResource):
   pass
 
 
+class DistrictAdmin(ListableAPIResource):
+  @classmethod
+  def class_url(cls):
+    return "/%s/district_admins" % API_VERSION
+
+
 class School(ListableAPIResource):
   pass
 
 
+class SchoolAdmin(ListableAPIResource):
+  @classmethod
+  def class_url(cls):
+    return "/%s/school_admins" % API_VERSION
+
+
 class Section(ListableAPIResource):
   pass
 
