@@ -3,6 +3,7 @@ import os
 import sys
 import unittest
 import httmock
+import itertools
 
 sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..'))
 import clever
@@ -21,32 +22,6 @@ def functional_test(auth):
       clever.api_base = os.environ.get('CLEVER_API_BASE', 'https://api.clever.com')
       if auth.get("token", None):
         clever.set_token(auth["token"])
-      elif auth.get("api_key", None):
-        clever.set_api_key(auth["api_key"])
-
-    def test_dns_failure(self):
-      api_base = clever.api_base
-      try:
-        clever.api_base = 'https://my-invalid-domain.ireallywontresolve/v1'
-        self.assertRaises(clever.APIConnectionError, clever.District.all)
-      finally:
-        clever.api_base = api_base
-
-    def test_list_accessors(self):
-      district = clever.District.all()[0]
-      self.assertEqual(district['name'], district.name)
-
-    def test_unicode(self):
-      # Make sure unicode requests can be sent
-      self.assertRaises(clever.InvalidRequestError, clever.District.retrieve, id=u'☃')
-
-    def test_none_values(self):
-      district = clever.District.all(sort=None)[0]
-      self.assertTrue(district.id)
-
-    def test_missing_id(self):
-      district = clever.District()
-      self.assertRaises(clever.InvalidRequestError, district.refresh)
   return FunctionalTests
 
 
@@ -55,8 +30,11 @@ class CleverTestCase(unittest.TestCase):
   def setUp(self):
     super(CleverTestCase, self).setUp()
     clever.api_base = os.environ.get('CLEVER_API_BASE', 'https://api.clever.com')
-    clever.set_api_key('DEMO_KEY')
+    clever.set_token('DEMO_TOKEN')
 
+#generates httmock responses for test_unicode_receive
+def unicode_content(url, request):
+  return {'status_code': 200, 'content': '{"data": {"name": "Oh haiô"}}'}
 
 class FunctionalTests(CleverTestCase):
 
@@ -76,17 +54,46 @@ class FunctionalTests(CleverTestCase):
     for district in clever.District.iter():
       self.assertTrue(district.id)
 
+  def test_starting_after(self):
+    allstudents = clever.Student.all()
+    second_to_last_id = allstudents[len(allstudents)-2]['id']
+    students = clever.Student.iter(starting_after=second_to_last_id)
+    count = len(list(students))
+    self.assertTrue(count==1)
+
+  def test_ending_before(self):
+    allstudents = clever.Student.all()
+    second_id = allstudents[1]['id']
+    students = clever.Student.iter(ending_before=second_id)
+    count = len(list(students))
+    self.assertTrue(count==1)
+
+  def test_iter_count(self):
+      r = requests.get('https://api.clever.com/v1.1/students?count=true',
+          headers={'Authorization': 'Bearer DEMO_TOKEN'})
+
+      req_count = json.loads(r.text)["count"]
+      iter_count = len([x for x in clever.Student.iter()])
+
+      self.assertTrue(req_count > clever.Student.ITER_LIMIT)
+      self.assertEqual(req_count, iter_count)
+
   def test_unsupported_params(self):
     self.assertRaises(clever.CleverError, lambda: clever.District.all(page=2))
     self.assertRaises(clever.CleverError, lambda: clever.District.all(limit=10))
     self.assertRaises(clever.CleverError, lambda: clever.District.all(page=2, limit=10))
 
-  def test_unicode(self):
-    # Make sure unicode requests can be sent
+  def test_unicode_send(self):
+    # Make sure unicode requests can be sent. 404 error is a clever.APIError
     self.assertRaises(clever.APIError, clever.District.retrieve, id=u'☃')
 
+  def test_unicode_receive(self):
+    with HTTMock(unicode_content):
+      # Make sure unicode responses can be received.
+      self.assertEqual(u'Oh haiô', clever.District.retrieve('something').name)
+
   def test_none_values(self):
-    district = clever.District.all(sort=None)[0]
+    district = clever.District.all(count=None)[0]
     self.assertTrue(district.id)
 
   def test_missing_id(self):
@@ -98,19 +105,23 @@ class FunctionalTests(CleverTestCase):
     self.assertEqual(clever_object.keys(), set())
     self.assertEqual(clever_object.values(), set())
 
+  def test_empty_list_on_no_data(self):
+    district = clever.District.all(where=json.dumps({'name': 'asdf'}))
+    self.assertEqual(district, [])
+
 class AuthenticationErrorTest(CleverTestCase):
 
   def test_invalid_credentials(self):
-    key = clever.get_api_key()
+    token = clever.get_token()
     try:
-      clever.set_api_key('invalid')
+      clever.set_token('invalid')
       clever.District.all()
     except clever.AuthenticationError, e:
       self.assertEqual(401, e.http_status)
       self.assertTrue(isinstance(e.http_body, str))
       self.assertTrue(isinstance(e.json_body, dict))
     finally:
-      clever.set_api_key(key)
+      clever.set_token(token)
 
 
 class InvalidRequestErrorTest(CleverTestCase):
@@ -145,11 +156,9 @@ class TooManyRequestsErrorTest(CleverTestCase):
 if __name__ == '__main__':
   suite = unittest.TestSuite()
   for TestClass in [
-          functional_test({"api_key": "DEMO_KEY"}),
-          functional_test({"token": "7f76343d50b9e956138169e8cbb4630bb887b18"}),
+          functional_test({"token": "DEMO_TOKEN"}),
           AuthenticationErrorTest,
           InvalidRequestErrorTest,
           TooManyRequestsErrorTest]:
     suite.addTest(unittest.TestLoader().loadTestsFromTestCase(TestClass))
   unittest.TextTestRunner(verbosity=2).run(suite)
-
